import pickle
import matplotlib.pyplot as plt
import numpy as np
import argparse


def parse_arguments():
    parser = argparse.ArgumentParser()
    
    parser.add_argument("-ID", nargs='+', help="experiment ID", type=str)
    parser.add_argument("--baseline", help="type of baseline to use.", 
                        type=str, nargs='+', default=None)
    parser.add_argument("--gradient", help="which archived gradient run to use.", 
                        type=str, default=None)
    parser.add_argument("--std_idx", help="iteration to print stdanard deviations for.", 
                        type=int, default=-1)
    args = parser.parse_args()
    
    return args

args = parse_arguments()

label_dict = {
    '2021-05-11_13-24-37a': 'RBF-RBF',
    '2021-05-11_13-24-37b': 'Sum',
    '2021-05-11_13-24-37c': 'RBF-Product',
    '2021-05-05_15-39-49_': 'RBF-RBF',
    '2021-05-06_15-49-09a': 'Sum',
    '2021-05-06_15-49-09b': 'RBF-Product',
    '2021-07-09_14-59-30_': 'RBF-RBF',
    '2021-07-27_18-28-38_': 'Sum',
    '2021-07-30_09-54-25_': 'RBF-Product',
    '2021-08-19_15-06-45a': 'RBF-Product',
    '2021-08-19_15-06-45b': 'Sum',
    '2021-08-19_15-06-45c': 'RBF-RBF'
}

colour_dict = {
    'RBF-Product': 'k',
    'Sum': 'k',
    'RBF-RBF': 'k',
    'Product grad': 'k',
    'Sum grad': 'k',
    'RBF grad': 'k',
    'Random': 'k',
    'Random_no_rep': 'k'
}

lstyle_dict = {
    'RBF-Product': '-',
    'Sum': '--',
    'RBF-RBF': ':',
    'Product grad': '-',
    'Sum grad': '--',
    'RBF grad': ':',
    'Random': '-.',
    'Random_no_rep': '-.'
}

label_disp_dict = {
    'RBF-Product': 'RBF-Product',
    'Sum': 'Sum',
    'RBF-RBF': 'RBF-RBF',
    'Product grad': r'Product $\dagger$',
    'Sum grad': r'Sum $\dagger$',
    'RBF grad': r'RBF $\dagger$',
    'Random': 'Random',
    'Random_no_rep': 'Random_no_rep'
}

std_dict = {
    'RBF-RBF': 0,
    'Sum': 1,
    'RBF-Product': 2,
    'grad RBF': 3,
    'grad Sum': 4,
    'grad Product': 5,
    'Random': 6,
    'Random no rep': 7
}

legend_order = {
    'RBF-Product': 0,
    'Sum': 1,
    'RBF-RBF': 2,
    r'Product $\dagger$': 3,
    r'Sum $\dagger$': 4,
    r'RBF $\dagger$': 5,
    'Random': 6,
    'Random no rep': 7
}

ratio_stds = ['']*8
dist_stds = ['']*8

baselines = {}
print(args.baseline)
if args.baseline is not None:
    for bb in args.baseline:
        print(bb)
        if bb in ["strong", "median", "weak"]:
            bl_file = '../Plume/baselines/baseline_%s_nd_norm2_log_P50_L100_Q1000.p' %bb
        elif bb in ["selection", "remaining_wts"]:
            bl_file = '../Plume/baselines/baseline_%s_nd_norm2_log_P100_L100_Q1000.p' %bb
        elif bb in ["selection_test"]:
            bl_file = '../Plume/baselines/baseline_selection_nd_test_norm2_log__P100_L100_Q1000.p' 
        elif bb in ["laqn"]:
            bl_file = '../Plume/baselines/laqn_baseline_temporal_log_mean_day-P194_L100_Q1000.p'
        elif bb in ["laqn_NO2_Roadside"]:
            bl_file = '../Plume/baselines/laqn_baseline_temporal_log_mean_NO2_Roadside_day-P214_L100_Q1000.p'
        elif bb in ["laqn_NO2_Roadside_2016"]:
            bl_file = '../Plume/baselines/laqn_baseline_temporal_log_mean_NO2_Roadside_day-2016_P365_L100_Q1000.p'
        elif bb in ["laqn_NO2_Roadside_2016_km"]:
            bl_file = '../Plume/baselines/laqn_baseline_spatial_log_mean_NO2_Roadside_km_tt_40_day-2016_P365_L100_Q1000_correct_no_rep.p'
        bl_dict = pickle.load(open(bl_file, 'rb'))
        baselines[bb] = {}
        baselines[bb]['dist'] = bl_dict['uniform']['dist']
        baselines[bb]['true_ratio'] = bl_dict['uniform']['ratio']
        if 'uniform_no_rep' in bl_dict.keys():
            bb_no_rep = bb+'_no_rep'
            baselines[bb_no_rep] = {}
            baselines[bb_no_rep]['dist'] = bl_dict['uniform_no_rep']['dist']
            baselines[bb_no_rep]['true_ratio'] = bl_dict['uniform_no_rep']['ratio']
    
    
if args.gradient == 'strong':
    ratio_grad = pickle.load(open('plume-values/values_[685056]_true.p', 'rb'))
    dist_grad = pickle.load(open('plume-values/values_[685056]_dist.p', 'rb'))    
elif args.gradient == 'selection':
    ratio_grad = pickle.load(open('plume-values/values_[685064]_true.p', 'rb'))
    dist_grad = pickle.load(open('plume-values/values_[685064]_dist.p', 'rb'))
    
num_plots = 2 
if args.baseline[0].split('_')[0] == 'laqn':
    fig, axs = plt.subplots(num_plots, 1, figsize=[3, 5])
else:
    fig, axs = plt.subplots(num_plots, 1, figsize=[3, 5])

ii = 0
for ID in args.ID:
    oview = open('preprocessed_runs/overview_%s.txt' %ID, 'r')
    line = oview.readline()

    cats = {}

    while len(line) > 0:
        try:
            ff, acq_str, sol_kernel, hyp_samples = line.split(', ')
            key = (acq_str, sol_kernel, hyp_samples)
        except:
            ff, acq_str, hyp_samples = line.split(', ')
            key = (acq_str, hyp_samples)
        hyp_samples = hyp_samples[:-1]

        file_name = ff.replace('/', '_')
        try:
            metrics = pickle.load(open('preprocessed_runs/%s_metrics.p' %file_name, 'rb'))
        except:
            metrics = pickle.load(open('preprocessed_runs/%s_%s_metrics.p' %(ID, file_name), 'rb'))
        for met in metrics.keys():
            if met not in cats.keys():
                cats[met] = {}
            if key not in cats[met].keys():
                cats[met][key] = []
            cats[met][key].append(metrics[met])
        line = oview.readline()



    ff = 0
    for met in cats.keys():
        pp = 0
        if num_plots > 1:
            ax = axs[ff]
        else:
            ax = axs
        for key in cats[met].keys():
            label = label_dict[key[2][:len('2021-05-11_13-24-37a')]]
            colour = colour_dict[label]
            linestyle = lstyle_dict[label]
            mean = np.mean(cats[met][key], 0)
            std = np.std(cats[met][key], 0)
            P, Q = np.shape(cats[met][key]) 
            
            mean_std_idx = mean[args.std_idx]
            std_std_idx = std[args.std_idx]/np.sqrt(P)
            if met == 'true_ratio':
                std_range = '%s-%s' %(format(mean_std_idx-std_std_idx, '.3f'), format(mean_std_idx+std_std_idx, '.3f'))
                ratio_stds[std_dict[label]] = std_range
            else:
                if args.baseline[0].split('_')[0] != 'laqn':
                    std_range = '%s-%s' %(format(7*(mean_std_idx-std_std_idx), '.3f'), format(7*(mean_std_idx+std_std_idx), '.3f'))
                else:
                    std_range = '%s-%s' %(format(mean_std_idx-std_std_idx, '.3f'), format(mean_std_idx+std_std_idx, '.3f'))
                dist_stds[std_dict[label]] = std_range
            print(met, label, std_range)
            
            ax.plot(range(1, Q+1), mean, color=colour, 
                    linestyle=linestyle, label=label)
            pp += 1
        



        ff += 1
    ii += pp
    

if args.gradient is not None:
    for kernel in ['RBF', 'Product', 'Sum']:
        kernel_str = '%s_ucb' %kernel
        label = '%s grad' %kernel
        colour = colour_dict[label]
        linestyle = lstyle_dict[label]
        mean_std_idx = ratio_grad[kernel_str][1][args.std_idx]
        std_std_idx = ratio_grad[kernel_str][2][args.std_idx]
        std_range = '%s-%s' %(format(mean_std_idx-std_std_idx, '.3f'), format(mean_std_idx+std_std_idx, '.3f'))
        print('true_ratio', kernel, std_range)
        ratio_stds[std_dict['grad %s' %kernel]] = std_range

        axs[0].plot(range(1, Q+1), ratio_grad[kernel_str][1], color=colour, 
                    linestyle=linestyle, label=label_disp_dict[label], alpha=0.4)

        mean_std_idx = dist_grad[kernel_str][1][args.std_idx]
        std_std_idx = dist_grad[kernel_str][2][args.std_idx]
        if args.baseline[0].split('_')[0] != 'laqn':
            std_range = '%s-%s' %(format(7*(mean_std_idx-std_std_idx), '.3f'), format(7*(mean_std_idx+std_std_idx), '.3f'))
        else:
            std_range = '%s-%s' %(format(mean_std_idx-std_std_idx, '.3f'), format(mean_std_idx+std_std_idx, '.3f'))
        print('dist', kernel, std_range)
        dist_stds[std_dict['grad %s' %kernel]] = std_range
        axs[1].plot(range(1, Q+1), dist_grad[kernel_str][1], color=colour, 
                    linestyle=linestyle, label=label_disp_dict[label], alpha=0.4)


bl_ii = 0
for met in cats.keys():
    ax = axs[bl_ii]
    alph = 0.75 
    for bb in baselines.keys():
        if met in baselines[bb].keys():
            mean = baselines[bb][met]['mean'][:Q]
            mean_std_idx = mean[args.std_idx]
            std_std_idx = baselines[bb][met]['mean_std'][args.std_idx]
            
            label = 'Random'
            if bb[-6:] == 'no_rep':
                label = 'Random no rep'    
                print('hello')
                if 'selection_test' in args.baseline:
                    continue
            if met == 'true_ratio':
                std_range = '%s-%s' %(format(mean_std_idx-std_std_idx, '.3f'), format(mean_std_idx+std_std_idx, '.3f'))
                ratio_stds[std_dict[label]] = std_range
            else:
                if args.baseline[0].split('_')[0] != 'laqn':
                    print('Not laqn')
                    std_range = '%s-%s' %(format(7*(mean_std_idx-std_std_idx), '.3f'), format(7*(mean_std_idx+std_std_idx), '.3f'))
                else:
                    std_range = '%s-%s' %(format(mean_std_idx-std_std_idx, '.3f'), format(mean_std_idx+std_std_idx, '.3f'))
                dist_stds[std_dict[label]] = std_range
            colour=colour_dict[label.replace(' ','_')]
            linestyle=lstyle_dict[label.replace(' ','_')]
            print(bb[-6:])

            ax.plot(range(1, Q+1), mean, color=colour, 
                linestyle=linestyle,
                label=label, 
                alpha=alph)
            alph = alph - 0.3
    bl_ii += 1



for ax in axs:
    ax.grid()
    ax.set_xlim(1, 100)
    if not args.baseline[0].split('_')[0] == 'laqn':
        ax.set_xscale('log')
        ax.set_xticklabels(['', 1, 10, 100, ''], size=10)

    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    
    
   
if args.baseline[0] in ['selection', 'selection_test']:
    axs[1].legend(loc='lower left', prop={"size":9}, facecolor='white', framealpha=1)
    handles, labels = axs[1].get_legend_handles_labels()
    labels, handles = zip(*sorted(zip(labels, handles), key=lambda t: legend_order[t[0]]))
    axs[1].legend(handles, labels, loc='lower left', prop={"size":9}, facecolor='white', framealpha=1)
else:
    axs[0].legend(loc='lower right', prop={"size":9}, facecolor='white', framealpha=1)
    handles, labels = axs[0].get_legend_handles_labels()
    labels, handles = zip(*sorted(zip(labels, handles), key=lambda t: legend_order[t[0]]))
    axs[0].legend(handles, labels, loc='lower right', prop={"size":9}, facecolor='white', framealpha=1)
    
    
axs[1].set_xlabel('Iteration')
axs[0].set_ylabel(r'Maximum ratio $\hat{y}/y^*$', size=10)
axs[1].set_ylabel(r'Distance $|\hat{x}-x^*|$ (km)', size=10)

ylim = axs[1].get_ylim()






if args.baseline[0].split('_')[0] == 'laqn':
    title = args.baseline[0].replace('_', ' ')
    title = 'London data'
    axs[0].set_title(title, size=10)
else:
    axs[0].set_title('%s subset' %args.baseline[0].replace('_', ' ').capitalize(), size=10)
    st_tick = ylim[0]*7//10*10+10    
    ticks = np.linspace(0, 100, 11)/7
    axs[1].set_yticks(ticks)
    axs[1].set_yticklabels(np.hstack(['', [int(tt) for tt in ticks[1:-1]*7], '']))
    
if ylim[0] < 0:
    ylim = (0, ylim[1])
axs[1].set_ylim(ylim)

ylim = axs[0].get_ylim()
ticks = np.linspace(0, 1, 11)
axs[0].set_yticks(ticks)
axs[0].set_ylim(ylim)
    
plt.savefig('SYR_figures/SYR_%s_%s_%s_bw.png' %(args.baseline, args.ID, args.gradient), bbox_inches='tight', ddpi=400, pad_inches=0)
plt.savefig('SYR_figures/SYR_%s_%s_%s_bw.svg' %(args.baseline, args.ID, args.gradient), bbox_inches='tight', ddpi=400, pad_inches=0)
if args.baseline[0] == 'selection':
    axs[1].legend().set_visible(False)
else:
    axs[0].legend().set_visible(False)
plt.savefig('SYR_figures/SYR_%s_%s_%s_bw_no_legend.svg' %(args.baseline, args.ID, args.gradient), bbox_inches='tight', ddpi=400, pad_inches=0)
plt.savefig('SYR_figures/SYR_%s_%s_%s_bw_no_legend.png' %(args.baseline, args.ID, args.gradient), bbox_inches='tight', ddpi=400, pad_inches=0)

print(std_dict)
print(ratio_stds)
print(dist_stds)

for name, builder in [('ratio', ratio_stds), ('dist', dist_stds)]:
    print(name)
    build = ''
    for ii in range(len(ratio_stds)):
        if ii == 3:
            print(build)
            build = ''
        elif ii != 0:
            build += ' & '
        build += builder[ii]
    print(build)
plt.show()
