print('Staring script to calculate baselines')

import numpy as np
import matplotlib.pyplot as plt
import scipy.stats as st

import copy
import time
import datetime
import pickle
import copy
import os

import sys
sys.path.append('../')

from source import kernel_functions_mul as kernel_functions
from source import setup_helper
from source import outer_loop as ol
from source import optimize_hyperparameters
from source import run_record

import sobol_seq


########## Parameters ###############
P = 100
Q = 1000
L = 100
year = 2016
prob_type = 'spatial_log_mean_NO2_Roadside_km_tt_40_day-' 

prob_folder = '../../LAQN/data_sorted/%s_problems/' %year
do_sobel = False
######################################



    

problems = []
lpt = len(prob_type)
files = [ff for ff in os.listdir(prob_folder) if ff[:lpt] == prob_type
        ]
print('Loading problems')
for ff in files:
    problems.append(pickle.load(open(prob_folder+ff, 'rb')))
P = len(problems)
print('Finished loading problems')

out_file = 'baselines/laqn_baseline_%s%s_P%s_L%s_Q%s_correct_no_rep.p' %(
    prob_type, year, P, L, Q)
print(out_file)

    
#### Simulate SOBL sampling
np.random.seed(13)

# sobl_seq_pre = sobol_seq.i4_sobol_generate(2, L)
sobl_seq_pre = sobol_seq.i4_sobol_generate(2, Q)*27
xx_all = np.array([[aa, bb] for aa in range(28) for bb in range(28)])

close_mat = []
for ii in range(Q):
    ss = sobl_seq_pre[ii]
    close_ii = sorted(xx_all, 
                      key=lambda aa: np.sqrt((aa[0]-ss[0])**2+(aa[1]-ss[1])**2))
    close_mat.append(close_ii[:80])

close_mat = np.array(close_mat)

def get_metrics(yy, xx, domain, true_func, true_maximum, true_maximiser):
    samp_max = np.max(yy)
    scaled_est_ratio = samp_max/true_maximum
    
    loc_est_max = xx[np.argmax(yy)]
    selected_idx = np.argmax(np.all(
                    domain == loc_est_max,
                    1)) 
    scaled_true_ratio = true_func[selected_idx]/true_maximum
    true_dist = np.sqrt(np.sum((
        domain[selected_idx] - true_maximiser
    )**2)) 
    return true_dist, scaled_true_ratio

def get_sobel_samp_idx(qq, prob, close_mat):
    for aa in range(len(close_mat)):
        test_pp = (close_mat[qq,aa,0], close_mat[qq,aa,1])
        attempt_idx = np.where((
            prob.domain == test_pp).all(axis=1))[0]
        if len(attempt_idx) > 0:
            return attempt_idx[0]

samp_sobel = lambda qq, domain: get_sobel_samp_idx(qq, domain, close_mat)
    
def sample_uniform(qq, prob):
    return np.random.randint(prob.dom_size)

    
def evaluate_once(P, Q, problems, samp_function, no_replacement=False):
    
    run_times = []
    max_kk_ratio = []
    max_kk_dist = []
    # For each problem
    for kk in range(P):
        max_qq_kk_ratio = []
        max_qq_kk_dist = []
        run_time_qq_kk = []

        xx = copy.deepcopy(problems[kk].xx)
        yy = copy.deepcopy(problems[kk].yy)

        true_dist, scaled_true_ratio = get_metrics(
            yy, xx, problems[kk].domain, problems[kk].true_func,
            problems[kk].maximum, problems[kk].maximiser)
        max_qq_kk_ratio.append(scaled_true_ratio)
        max_qq_kk_dist.append(true_dist)
        
        if no_replacement:
            cand_idx = np.array(list(range(problems[kk].dom_size)))
            used = [np.argmax(np.sum(problems[kk].xx[ii] == problems[kk].domain, 1) == 2) 
                    for ii in range(len(problems[kk].xx))]
            rand_order = [ii for ii in cand_idx if ii not in used]
            np.random.shuffle(rand_order)

        # For each iteration (cf. Bayesian Optimisation)
        for qq in range(Q):

            start_time = time.time()
            if no_replacement:
                samp_idx = rand_order[qq%len(rand_order)]
            else:
                samp_idx = samp_function(qq, problems[kk])
            

            xx = np.vstack([xx, problems[kk].domain[samp_idx]])
            yy = np.hstack([yy,
                        problems[kk].true_func[samp_idx]
                        + np.random.normal(
                            0, np.sqrt(problems[kk].sampling_noise))])

            true_dist, scaled_true_ratio = get_metrics(
                yy, xx, problems[kk].domain, problems[kk].true_func,
                problems[kk].maximum, problems[kk].maximiser)
            max_qq_kk_ratio.append(scaled_true_ratio)
            max_qq_kk_dist.append(true_dist)

            run_time_qq_kk.append(time.time()-start_time)

        max_kk_ratio.append(max_qq_kk_ratio)
        max_kk_dist.append(max_qq_kk_dist)
        run_times.append(run_time_qq_kk)

    # Convert to arrays
    metrics_raw = {'ratio': np.array(max_kk_ratio),
                   'dist':  np.array(max_kk_dist),
                   'duration': np.array(run_times)}
    return metrics_raw

all_metrics = {}
if do_sobel:
    versions = [(samp_sobel, 'sobel', 1, False), (sample_uniform, 'uniform', L, False)]
else:
    versions = [(sample_uniform, 'uniform_no_rep', L, True), (sample_uniform, 'uniform', L, False)]
for sampler, label, num_runs, rep in versions:
    all_runs = []
    for ii in range(num_runs):
        print('Staring %s' %ii)
        with open('tmp_bl.txt', 'a') as out_ff:
            out_ff.write('Starting %s\n' %ii)
        all_runs.append(evaluate_once(P, Q, problems, sampler, rep))
        print('Finished %s ' %ii)
    
    metrics_raw = {}
    for run in all_runs:
        for mm in run.keys():
            if mm not in metrics_raw.keys():
                metrics_raw[mm] = np.array(run[mm])
            else:
                metrics_raw[mm] = np.vstack([metrics_raw[mm], run[mm]])
    print(label)
    metrics_label = {}
    for mm in metrics_raw.keys():
        print(np.shape(metrics_raw[mm]))
        metrics_label[mm] = {}
        mm_mean = np.mean(metrics_raw[mm], 0)
        mm_mean_std = np.std(metrics_raw[mm], 0)/np.sqrt(P*num_runs)
        metrics_label[mm]['mean'] = mm_mean
        metrics_label[mm]['mean_std'] = mm_mean_std
        metrics_label[mm]['Q'] = Q
    all_metrics[label] = metrics_label


pickle.dump(all_metrics, open(out_file, 'wb'))
