import matplotlib.pyplot as plt
import numpy as np
import pickle

import sys
sys.path.append('../')

from source import setup_helper


def generate_problem(data, M, kk, dd, name, return_prob=False,
                    mean_to_subtract=0, std_to_divide=1, take_log=False):
    sampling_noise = 10e-7
    np.random.seed(kk)
    X = []
    Y = []
    for ii in range(dd):
        for jj in range(dd):
            if data[kk,ii,jj] < np.inf:
                X.append([ii, jj])
                Y.append(data[kk,ii,jj])
    X = np.array(X)
    if take_log:
        Y = (np.log(np.array(Y))-mean_to_subtract)/std_to_divide
    else:
        Y = (np.array(Y)-mean_to_subtract)/std_to_divide

    N = len(Y)
    
    indices = list(range(N))
    np.random.shuffle(indices)
    samp_idx = indices[:M]

    xx_nn = X[samp_idx]
    yy_nn = Y[samp_idx] + np.random.normal(0, np.sqrt(sampling_noise), size=M)

    prob_kk = setup_helper.Problem_obj(
            xx_nn = xx_nn,
            yy_nn = yy_nn,
            domain = X,
            true_func = Y,
            kernel = None,
            hyp_dict = None,
            C = None,
            dom_1D_vecs = np.array([np.array(range(dd))]*2),
            seed=kk,
            sampling_noise=sampling_noise
        )
    
    pickle.dump(prob_kk, open('problems/%s.p' %name, 'wb'))
    if return_prob:
        return prob_kk

data = pickle.load(open('plume_raw.p', 'rb'))
to_ignore = pickle.load(open('duplicate_ignore.p', 'rb'))

D0, D1, D2 = np.shape(data)
data = data*1000

max_values = []
coverage = []
for ii in range(D0):
    non_nan = D1*D2 - np.sum(np.isnan(data[ii,:,:]))
    if non_nan == 0:
        print('%s does not contain any values' %ii)
    max_values.append(np.nanmax(data[ii,:,:]))
    coverage.append(non_nan)

indices = [ii for ii in range(D0) if coverage[ii] > 0.90*D1*D2 
           and ii not in to_ignore]

sorted_ind = sorted(indices, key=lambda jj: max_values[jj])


##############################################################################
### Start generating problems ################################################
##############################################################################
dd = 28
M = 10
# Generate problems, no weighting

for kk in range(100):
    generate_problem(data, M, kk, dd, 'plume_prob_%s' %kk)
    
strong_list = sorted_ind[-50:]
strong_list.reverse()
strong_tuning_list = sorted_ind[-60:-50]
strong_tuning_list.reverse()
weak_list = sorted_ind[:50]
weak_tuning_list = sorted_ind[50:60]
median_ind = sorted_ind[len(sorted_ind)//2-25:len(sorted_ind)//2+25]
median_tune_idx_list = sorted_ind[len(sorted_ind)//2-35:len(sorted_ind)//2-25]

prob_type_names_all = ['strong_nd', 'weak_nd', 'median_nd']
prob_type_list_all = [strong_list, weak_list, median_ind]
prob_type_tuning_list_all = [strong_tuning_list, weak_tuning_list, 
                         median_tune_idx_list]
remaining = [aa for aa in sorted_ind if 
                 aa not in np.hstack(prob_type_list_all)
                and aa not in np.hstack(prob_type_tuning_list_all)]

remaining_sel = np.array(remaining)[np.array(range(8, len(remaining), 9))]
remaining_sel_tune = np.array(remaining)[np.array(range(4, len(remaining), 9))]
remaining_sel_test = np.array(remaining)[np.array(range(7, len(remaining), 9))]
pickle.dump(remaining_sel_test, open('problems/%s_meta_data.p' %'remaining_wts_nd_test', 'wb'))

print(len(remaining_sel_tune))
print(len([aa for aa in remaining_sel if aa in remaining_sel_tune]))

prob_type_names_all.append('remaining_wts_nd')
prob_type_list_all.append(remaining_sel)
prob_type_tuning_list_all.append(remaining_sel_tune)

    
for ii in range(len(prob_type_names_all)):
    prob_type = prob_type_names_all[ii]
    prob_type_list = prob_type_list_all[ii]
    prob_type_tuning_list = prob_type_tuning_list_all[ii]
    
    for aa in range(len(prob_type_list)):
        kk = prob_type_list[aa]
        generate_problem(data, M, kk, dd, 'plume_prob_%s_%s' %(prob_type, aa))
    pickle.dump(prob_type_list, open('problems/%s_meta_data.p' %prob_type, 'wb'))

    tuning_problems = []
    for aa in range(len(prob_type_tuning_list)):
        kk = prob_type_tuning_list[aa]
        tuning_problems.append(
            generate_problem(data, M, kk, dd, 
                             'plume_prob_%s_tune_%s' %(prob_type, aa),
                            return_prob=True))
    pickle.dump(prob_type_tuning_list, 
                open('problems/%s_tune_meta_data.p' %prob_type, 'wb'))

    all_values = []
    all_values_lists = []
    for aa in range(len(tuning_problems)):
        all_values = np.hstack([all_values, tuning_problems[aa].true_func])
        all_values_lists.append(tuning_problems[aa].true_func)
    mean_to_subtract = np.mean(all_values)
    std_all = np.mean([np.std(ll) for ll in all_values_lists])
    pickle.dump(mean_to_subtract, open('%s_norm_mean.p' %prob_type, 'wb'))
    pickle.dump(std_all, open('%s_norm_std.p' %prob_type, 'wb'))

    for aa in range(len(prob_type_tuning_list)):
        kk = prob_type_tuning_list[aa]
        generate_problem(data, M, kk, dd, 
                         'plume_prob_%s_norm2_tune_%s' %(prob_type, aa),
                        mean_to_subtract=mean_to_subtract,
                        std_to_divide=std_all)
        
    for aa in range(len(prob_type_list)):
        kk = prob_type_list[aa]
        generate_problem(data, M, kk, dd, 
                         'plume_prob_%s_norm2_%s' %(prob_type, aa),
                        mean_to_subtract=mean_to_subtract,
                        std_to_divide=std_all)
        
    mean_to_subtract_log = np.mean(np.log(all_values))
    std_all_log = np.mean([np.std(np.log(ll)) for ll in all_values_lists])
    pickle.dump(mean_to_subtract_log, open('%s_norm_mean_log.p' %prob_type, 'wb'))
    pickle.dump(std_all_log, open('%s_norm_std_log.p' %prob_type, 'wb'))
    
    for aa in range(len(prob_type_tuning_list)):
        kk = prob_type_tuning_list[aa]
        generate_problem(data, M, kk, dd, 
                         'plume_prob_%s_norm2_log_tune_%s' %(prob_type, aa),
                        mean_to_subtract=mean_to_subtract_log,
                        std_to_divide=std_all_log, take_log=True)
        
    for aa in range(len(prob_type_list)):
        kk = prob_type_list[aa]
        generate_problem(data, M, kk, dd, 
                         'plume_prob_%s_norm2_log_%s' %(prob_type, aa),
                        mean_to_subtract=mean_to_subtract_log,
                        std_to_divide=std_all_log, take_log=True)
        
prob_type = 'median_nd'
mean_to_subtract_log = pickle.load(open('%s_norm_mean_log.p' %prob_type, 'rb'))
std_all_log = pickle.load(open('%s_norm_std_log.p' %prob_type, 'rb'))

prob_type_list = remaining
prob_type = 'remaining'

for aa in range(len(prob_type_list)):
    kk = prob_type_list[aa]
    generate_problem(data, M, kk, dd, 
                     'plume_prob_%s_norm2_log_%s' %(prob_type, aa),
                        mean_to_subtract=mean_to_subtract_log,
                        std_to_divide=std_all_log, take_log=True)
    
prob_type_list = remaining_sel
prob_type = 'selection_nd'

for aa in range(len(remaining_sel)):
    kk = prob_type_list[aa]
    generate_problem(data, M, kk, dd, 
                     'plume_prob_%s_norm2_log_%s' %(prob_type, aa),
                        mean_to_subtract=mean_to_subtract_log,
                        std_to_divide=std_all_log, take_log=True)
    
prob_type_list = remaining_sel_test
prob_type = 'selection_nd_test'
for aa in range(len(prob_type_list)):
    kk = prob_type_list[aa]
    generate_problem(data, M, kk, dd, 
                     'plume_prob_%s_norm2_log_%s' %(prob_type, aa),
                        mean_to_subtract=mean_to_subtract_log,
                        std_to_divide=std_all_log, take_log=True)
