#!/bin/sh
#SBATCH -N 1                # Nodes requested
#SBATCH -n 1                # Tasks requested
#SBATCH --cpus-per-task=8   # CPUs requested
#SBATCH --mem=14000         # Memory in Mb
#SBATCH --time=0-12:00:00    # Time required
#SBATCH --exclude=REDACTED
#SBATCH -o ./Report/slurm-%A_%a.out

export STUDENT_ID=$(whoami)

export PYTHON_PATH=$PATH

mkdir -p /disk/scratch/${STUDENT_ID}

JOBTASK="${SLURM_ARRAY_JOB_ID}-${SLURM_ARRAY_TASK_ID}"
echo ${JOBTASK}
echo ${1}
echo ${2}

export TMPDIR=/disk/scratch/${STUDENT_ID}/
export TMP=/disk/scratch/${STUDENT_ID}
export RESDIR=/disk/scratch/${STUDENT_ID}/${JOBTASK}

DATADIR=/home/${STUDENT_ID}/BO/$2 
SAMPLESDIR=/home/${STUDENT_ID}/MCMC/Hierarchical/Hyp_samples/

source /home/${STUDENT_ID}/miniconda3/bin/activate CONDA_ENVIRONMENT

echo "Machine:"
hostname
echo "Number of CPUS: "${SLURM_CPUS_PER_TASK}
TIMESTAMP=`date +%Y-%m-%d_%H-%M-%S`
echo "Timestamp: "${TIMESTAMP}
git log -1
git branch
conda info | grep active

experiment_text_file=$1
echo ${experiment_text_file}
echo ${SLURM_ARRAY_JOB_ID}
echo ${SLURM_ARRAY_TASK_ID}
COMMAND="`sed \"${SLURM_ARRAY_TASK_ID}q;d\" ${experiment_text_file}`"
echo ${COMMAND}

prob_file_raw=$(echo $COMMAND | grep -o -P '(?<=prob_file ).*(?=-)')
if [ $? -eq 0 ]; then
    prob_file=`echo $prob_file_raw|cut -d ' ' -f1`
    echo "Prob file: "${DATADIR}${prob_file}
    echo "Transferring from ${DATADIR}${prob_file} to ${TMPDIR}${prob_file}"
    rsync -avzh ${DATADIR}${prob_file} ${TMPDIR}${prob_file}
else
    prob_file_raw=$(echo $COMMAND | grep -o -P '(?<=prob_file ).*')
    if [ $? -eq 0 ]; then
        prob_file=`echo $prob_file_raw|cut -d ' ' -f1`
        echo "Prob file: "${DATADIR}${prob_file}
        echo "Transferring from ${DATADIR}${prob_file} to ${TMPDIR}${prob_file}"
        rsync -avzh ${DATADIR}${prob_file} ${TMPDIR}${prob_file}
    fi
fi

DATADIR_NODE=/disk/scratch/${STUDENT_ID}/DATA/
samples_file_raw=$(echo $COMMAND | grep -o -P '(?<=hyp_samples_file ).*(?=-)')
if [ $? -eq 0 ]; then
    samples_file=`echo $samples_file_raw|cut -d ' ' -f1`
    echo "Samples file: "${SAMPLESDIR}${samples_file}
    echo "Transferring from ${SAMPLESDIR}${samples_file} to ${DATADIR_NODE}${samples_file}"
    mkdir ${DATADIR_NODE}
    rsync -avzh ${SAMPLESDIR}${samples_file} ${DATADIR_NODE}${samples_file}
else
    samples_file_raw=$(echo $COMMAND | grep -o -P '(?<=hyp_samples_file ).*')
    if [ $? -eq 0 ]; then
        samples_file=`echo $samples_file_raw|cut -d ' ' -f1`
        echo "Samples file: "${SAMPLESDIR}${samples_file}
        echo "Transferring from ${SAMPLESDIR}${samples_file} to ${DATADIR_NODE}${samples_file}"
        mkdir ${DATADIR_NODE}
        rsync -avzh ${SAMPLESDIR}${samples_file} ${DATADIR_NODE}${samples_file}
    fi
fi



COMMAND_FULL="${COMMAND} ${JOBTASK}"

echo "Running provided command: ${COMMAND_FULL}"
eval "${COMMAND_FULL}"
echo "Command ran successfully!"

cd ${RESDIR}
if [ $? -eq 0 ]; then
    zip -rDo ${TMPDIR}${JOBTASK}.zip *

    OUT_FOLDER=/home/${STUDENT_ID}/BO/Outputs/${SLURM_ARRAY_JOB_ID}
    echo "Sending results to: ${OUT_FOLDER}/${JOBTASK}"
    mkdir -p ${OUT_FOLDER}

    rsync -avzh ${TMPDIR}${JOBTASK}.zip ${OUT_FOLDER}/${JOBTASK}.zip
fi


e_file=${JOBTASK}_email.txt
exp_id=${SLURM_ARRAY_JOB_ID}
echo "Subject: Experiment ${exp_id} finished" > ${e_file}
echo "The experiment has finished running on the cluster." >> ${e_file}
date >> ${e_file}
cat ${e_file}
echo ${exp_id} >> ${e_file}
if [ "$(squeue | grep -c ${exp_id})" -ge 2 ]
then
    echo 'Still running'
else
    echo 'Experiment finished'
    sendmail REDACTED < ${e_file}
fi
