"""Module holding class to store GP hyperparameters"""
import copy
import autograd.numpy as np

class Val_holder:
    pass

def convert_kernel_type(old_hyp_dict, new_type):
    tmp_hyp_dict = {"kernel_name": new_type}
    for key in old_hyp_dict.key_dict.keys():
        tmp_hyp_dict[key] = \
            old_hyp_dict.array[old_hyp_dict.key_dict[key]]
    if old_hyp_dict == 'Product' and new_type == 'RBF':
        tmp_hyp_dict['sigma_rbf'] = old_hyp_dict.array[old_hyp_dict.key_dict['sigma_hat']]
    
    if old_hyp_dict == 'RBF' and new_type == 'Product':
        tmp_hyp_dict['sigma_hat'] = old_hyp_dict.array[old_hyp_dict.key_dict['sigma_rbf']]
    return Hyp_dict(tmp_hyp_dict)

def get_hyp_dict(name, parts=None):
    if name in ['RBF', 'RBF_ARD', 'Directed', 'Sum', 'Product']:
        return Hyp_dict({'kernel_name': name})
    elif name == 'Combination':
        key_dict = {}
        ii = 0
        for pp in parts:
            part_dict = get_hyp_dict(pp)
            key_dict[pp] = {}
            for key in part_dict.key_dict.keys():
                key_dict[pp][key] = ii + part_dict.key_dict[key]
            ii += part_dict.length
        self.key_dict = key_dict
        self.prior_dict = None 
        self.length = ii
    else:
        raise ValueError

class Hyp_dict:
    """Class to store GP hyperparameters"""

    def combine_parts(self, parts, way):
        assert way == 'Summation'
        
        key_dicts = []
        pp_num = 0
        ii = 1
        constant = {}
        linking = {'sigma_2': 0}
        try:
            both_array_list = [parts[0].array[
                parts[0].key_dict['sigma_2']]]
        except:
            both_array_list = [parts[0].array[
                parts[0].key_dict[0]['sigma_2']]]
        values_both = []
        for part_dict in parts:
            
            if part_dict.name in ['Combination', 'Temporal']:
                key_dicts.append([])
                values_both.append(Val_holder())
                values_both[-1].values = part_dict.values
                
                for rr_num in range(len(part_dict.key_dict)):
                    key_dicts[pp_num].append({})
                    for key in part_dict.key_dict[rr_num].keys():
                        if key in constant.keys():
                            values_both.values[pp_num].values[rr_num].__dict__[key] = 1
                        elif key in linking.keys():
                            key_dicts[pp_num][rr_num][key] =\
                                linking[key]
                        else:
                            kk_idx = part_dict.key_dict[rr_num][key]
                            key_dicts[pp_num][rr_num][key] = ii
                            both_array_list.append(
                                part_dict.array[kk_idx])
                            ii += 1

            else:
                key_dicts.append({})
                cur_vals = Val_holder()
                for val_key in part_dict.__dict__.keys():
                    if val_key in [
                        'sigma_2', 'sigma_rbf', 'sigma_hat',
                    'sigma_dir', 'll', 'll_dir', 'direc']:
                        cur_vals.__dict__[val_key] = \
                            part_dict.__dict__[val_key]
                values_both.append(cur_vals)
                
                for key in part_dict.key_dict.keys():
                    if key in constant.keys():
                        cur_vals.__dict__[key] = 1
                    elif key in linking.keys():
                        key_dicts[pp_num][key] = linking[key]
                    else:
                        kk_idx = part_dict.key_dict[key]
                        key_dicts[pp_num][key] = \
                            ii
                        both_array_list.append(
                            part_dict.array[kk_idx])
                        ii += 1
            pp_num += 1

        D_list, parts_list, ll_lengths = [], [], []
        for hh in parts:
            if hh.name in ['Combination', 'Temporal']:
                D_list.append(hh.D_list)
                parts_list.append(hh.parts)
                ll_lengths.append(hh.ll_lengths)
            else:
                D_list.append(hh.D)
                parts_list.append(hh.name)
                ll_lengths.append(hh.ll_length)
        self.D_list = D_list
        self.parts = parts_list
        self.ll_lengths = ll_lengths
        self.way = way
        self.array = np.array(both_array_list)
        self.length = len(self.array)
        self.key_dict = key_dicts
        self.values = values_both
        self.update_vals_from_array(self.array)
    
    def __init__(self, hyp_normal_dict, parts=None, way=None, 
                 D_list=[2]):
        self.name = hyp_normal_dict["kernel_name"]
        if self.name not in ['Combination', 'Temporal']:
            self.D = D_list[0]
        
        if self.name in ['Combination', 'Temporal']:
            if type(parts[0]) is not str:
                self.combine_parts(parts, way)
                return
            self.parts = parts
            self.values = []
            self.way = way
            self.D_list = D_list
            key_dicts = []
            ll_lengths = []
            ii = 0
            pp_num = 0
            for pp in parts:
                part_dict = Hyp_dict({'kernel_name': pp})
                
                cur_vals = Val_holder()
                if pp_num == 0:
                    linking = {}
                else:
                    linking = {'sigma_2': 
                              key_dicts[0]['sigma_2']}
                if self.way == 'Summation':
                    constant = {}
                elif self.way == 'Multiplication':
                    if pp == 'Sum':
                        constant = {}
                    elif pp_num == 0 and 'Sum' not in parts:
                        constant = {}
                    else:
                        cons = ['sigma_rbf', 'sigma_hat']
                        constant = {cc: part_dict.key_dict[cc] 
                                    for cc in cons 
                                    if cc in part_dict.key_dict.keys()}
                    
                ll_lengths.append(part_dict.ll_length)
                key_dicts.append({})
                
                for key in part_dict.key_dict.keys():
                    if key in constant.keys():
                        cur_vals.__dict__[key] = 1
                    elif key in linking.keys():
                        key_dicts[pp_num][key] = \
                            linking[key]
                    else:
                        kk_idx = part_dict.key_dict[key]
                        link_vals = np.array(
                            [linking[vv] for vv in linking.keys() 
                             if vv in part_dict.key_dict.keys()])
                        cons_vals = np.array(
                            [constant[vv] for vv in constant.keys()
                             if vv in part_dict.key_dict.keys()])
                        key_dicts[pp_num][key] = \
                            ii + kk_idx \
                            - np.sum(kk_idx > link_vals) \
                            - np.sum(kk_idx > cons_vals)
                        
                assert(key_dicts[pp_num]["sigma_2"] == 0)
                self.values.append(cur_vals)
                ii += part_dict.length - len(linking) - len(constant)
                pp_num += 1
            self.key_dict = key_dicts

            self.prior_dict = None
            self.length = ii
            self.ll_lengths = ll_lengths
            

            
        elif self.name == "RBF":  # RBF kernel
            self.key_dict = {
                "sigma_2": 0,
                "sigma_rbf": 1,
                "ll": 2,
            }
            self.prior_dict = self.key_dict
            self.length = 3
            self.ll_length = 1
            
        elif self.name == "RBF_ARD":  # RBF ARD kernel
            self.key_dict = {
                "sigma_2": 0,
                "sigma_rbf": 1,
                "ll": 2,
            }
            self.prior_dict = self.key_dict
            self.length = 2 + len(hyp_normal_dict["ll"])
            self.ll_length = len(hyp_normal_dict["ll"])

        elif self.name == "Directed":  # Directed kernel
            self.key_dict = {
                "sigma_2": 0,
                "direc": 1,
                "sigma_dir": 2,
                "ll_dir": 3,
            }
            self.length = 4
            self.ll_length = 0

        elif self.name == "Sum":  # Wind sum kernel
            self.key_dict = {
                "sigma_2": 0,
                "direc": 1,
                "sigma_dir": 2,
                "ll_dir": 3,
                "sigma_rbf": 4,
                "ll": 5,
            }
            self.length = 6
            self.ll_length = 1

        elif self.name == "Product": # Wind product kernel
            self.key_dict = {
                "sigma_2": 0,
                "direc": 1,
                "sigma_hat": 2,
                "ll_dir": 3,
                "ll": 4
            }
            self.length = 5
            self.ll_length = 1
        else:
            raise ValueError

        if self.name in ['Directed', 'Sum', 'Product']:
            trunc_mat = np.zeros((self.length, self.length-1))
            trunc_mat[0,0] = 1
            for ii in range(2, self.length):
                trunc_mat[ii,ii-1] = 1
            self.trunc_mat = trunc_mat
            
            self.prior_dict = {}
            for key in self.key_dict.keys():
                if key == "direc":
                    continue
                if self.key_dict[key] > self.key_dict["direc"]:
                    self.prior_dict[key] = self.key_dict[key] - 1
                else:
                    self.prior_dict[key] = self.key_dict[key]
        self.array = np.ones(self.length)
        self.update_vals_from_dict(hyp_normal_dict)

    def __str__(self):
        if self.name in ['Combination', 'Temporal']:
            pass
        else:
            out_str = 'kernel: \t%s' %self.name
            for key in ['sigma_2', 'direc', 'll', 'll_dir', 
                        'sigma_rbf', 'sigma_dir', 'sigma_hat',
                        'array']:
                if key in self.__dict__.keys():
                    out_str = '%s\n%s: \t%s' %(
                        out_str, key, self.__dict__[key])
            return out_str

    def trunc_array(self, arr_to_trunc):
        if self.name in ['Combination', 'Temporal']:
            raise ValueError
        if self.name in ["RBF", "RBF_ARD"]:
            return arr_to_trunc
        elif self.name in ['Directed', 'Sum', 'Product']:
            return np.matmul(np.array(arr_to_trunc, ndmin=2), self.trunc_mat)
    
    def update_vals_from_dict_combination(self, hyp_normal_dict):
        """ Update the hyperparameters with values from 
        a normal dict"""
        input_keys = hyp_normal_dict.keys()
        for key in input_keys:
            splt = key.find('_')
            kk_str = key[:splt]
            key_part = key[splt+1:]
            if not kk_str.isnumeric():
                continue
            kk = int(kk_str)
            if type(self.key_dict[kk]) is dict:
                key_dict = self.key_dict[kk]
                part = self.parts[kk]
            else:
                splt_two = key_part.find('_')
                jj_str = key_part[:splt_two]
                key_part = key_part[splt_two+1:]
                if not jj_str.isnumeric():
                    if '%s_%s' %(jj_str, key_part) == 'sigma_2':
                        jj_str = '0'
                    else:
                        continue
                jj = int(jj_str)
                key_dict = self.key_dict[kk][jj]
                part = self.parts[kk][jj]
            if key_part in key_dict.keys():
                if key_part == 'll':
                    if part in {'Sum', 'RBF', 'Product', 'Directed'}:
                        try:
                            self.array[
                                key_dict[key_part]
                            ] = hyp_normal_dict[key][0]
                        except (IndexError, TypeError):
                            self.array[
                                key_dict[key_part]
                            ] = hyp_normal_dict[key]
                    else: # RBF_ARD
                        self.array[
                            key_dict[key_part]:
                            key_dict[key_part]\
                                +len(hyp_normal_dict[key])
                        ] = hyp_normal_dict[key]
                elif key_part == "direc":
                    try:
                        self.array[key_dict[key_part]] = \
                            np.arctan(
                                hyp_normal_dict[key][1]
                                / hyp_normal_dict[key][0]
                            )
                    except (IndexError, TypeError):
                        self.array[
                                key_dict[key_part]
                            ] = hyp_normal_dict[key]
                else:
                    self.array[
                        key_dict[key_part]
                    ] = hyp_normal_dict[key]
                    
        self.update_vals_from_array(self.array)
        
    def update_vals_from_dict(self, hyp_normal_dict):
        """ Update the hyperparameters with values from 
        a normal dict"""
        if self.name in ['Combination', 'Temporal']:
            self.update_vals_from_dict_combination(
                hyp_normal_dict)
        else:
            input_keys = hyp_normal_dict.keys()
            for key in input_keys:
                if key in self.key_dict.keys():
                    if key == "ll":
                        if self.name in {'Sum', 'RBF', 'Product'}:
                            try:
                                self.array[
                                    self.key_dict[key]
                                ] = hyp_normal_dict[key][0]
                            except (IndexError, TypeError):
                                self.array[
                                    self.key_dict[key]
                                ] = hyp_normal_dict[key]
                        else: # RBF_ARD
                            self.array[
                                self.key_dict[key]:self.length
                            ] = hyp_normal_dict[key]
                    elif key == "direc":
                        try:
                            self.array[self.key_dict[key]] = np.arctan(
                                hyp_normal_dict[key][1]
                                / hyp_normal_dict[key][0]
                            )
                        except (IndexError, TypeError):
                            self.array[
                                self.key_dict[key]
                            ] = hyp_normal_dict[key]
                    else:
                        self.array[
                            self.key_dict[key]
                        ] = hyp_normal_dict[key]

            self.update_vals_from_array(self.array)

    def update_vals_from_array_combination(self, hyp_array):
        self.array = copy.deepcopy(hyp_array)
        if 'values' not in self.__dict__.keys():
            print('no values found')
            self.values = []
        try:
            self.sigma_2 = self.array[self.key_dict[0]["sigma_2"]]
        except:
            self.sigma_2 = self.array[self.key_dict[0][0]["sigma_2"]]
        pp_idx = 0
        for pp_lst in self.parts:
            if type(pp_lst) is list:
                sub_list = [(pp_lst[rr], rr) 
                            for rr in range(len(pp_lst))]
                if len(self.values) <= pp_idx:
                    self.values.append(Val_holder())
                    self.values[pp_idx].values = []
                    for rr_idx in range(len(pp_lst)):
                        self.values[pp_idx].values.append(
                            Val_holder())
            else:
                sub_list = [(pp_lst, None)]
                if len(self.values) <= pp_idx:
                    self.values.append(Val_holder())
                    
            for pp, rr_idx in sub_list:
                if rr_idx is None:
                    pp_key_dict = self.key_dict[pp_idx]
                    cur_D = self.D_list[pp_idx]
                    cur_vals = self.values[pp_idx]
                    cur_ll_len = self.ll_lengths[pp_idx]
                else:
                    pp_key_dict = self.key_dict[pp_idx][rr_idx]
                    cur_D = self.D_list[pp_idx][rr_idx]
                    cur_vals = self.values[pp_idx].values[rr_idx]
                    cur_ll_len = self.ll_lengths[pp_idx][rr_idx]
                cur_vals.sigma_2 = self.array[pp_key_dict["sigma_2"]]

                if pp in {'Sum', 'RBF', 'Product'}:
                    cur_vals.ll = np.array(
                        [self.array[pp_key_dict["ll"]]] \
                        * cur_D)
                if pp in {"RBF", "Sum", "RBF_ARD"} \
                        and "sigma_rbf" in pp_key_dict.keys():
                    cur_vals.sigma_rbf = self.array[
                        pp_key_dict["sigma_rbf"]
                    ]
                if pp == "RBF_ARD":
                    cur_vals.ll = np.array(
                        self.array[
                            pp_key_dict["ll"]:
                            pp_key_dict["ll"] + cur_ll_len]
                    )
                if pp == 'Product' \
                        and "sigma_hat" in pp_key_dict.keys():
                    cur_vals.sigma_hat = self.array[
                        pp_key_dict["sigma_hat"]
                    ]
                if pp in {'Sum', 'Directed', 'Product'}:
                    gamma = self.array[pp_key_dict["direc"]]
                    cur_vals.direc = np.array([np.cos(gamma), 
                                               np.sin(gamma)])
                    if pp != 'Product':
                        cur_vals.sigma_dir = self.array[
                            pp_key_dict["sigma_dir"]
                        ]
                    cur_vals.ll_dir = self.array[pp_key_dict["ll_dir"]]
                    cur_vals.B_0 = 1 - cur_vals.direc[0] ** 2
                    cur_vals.B_1 = 1 - cur_vals.direc[1] ** 2
                    cur_vals.B_2 = -cur_vals.direc[0] * cur_vals.direc[1]
                    cur_vals.A = np.array(
                        [[cur_vals.B_0, cur_vals.B_2], 
                         [cur_vals.B_2, cur_vals.B_1]]
                    )
                if pp == 'Product':
                    ll_ratio = cur_vals.ll_dir**2/cur_vals.ll[0]**2
                    cur_vals.E_0 = ll_ratio + cur_vals.B_0
                    cur_vals.E_1 = ll_ratio + cur_vals.B_1
                    cur_vals.E_2 = cur_vals.B_2
                    cur_vals.E = np.array(
                        [[cur_vals.E_0, cur_vals.E_2], 
                         [cur_vals.E_2, cur_vals.E_1]]
                    )

                if rr_idx is None:
                    self.values[pp_idx] = cur_vals
                else:
                    self.values[pp_idx].values[rr_idx] =cur_vals

            pp_idx += 1
        
    def update_vals_from_array(self, hyp_array):
        """ Update the hyperparameters with values from an array"""
        # Set stored values equal to array values
        if self.name in ['Combination', 'Temporal']:
            self.update_vals_from_array_combination(hyp_array)
        else:
            self.array = copy.deepcopy(hyp_array)

            self.sigma_2 = self.array[self.key_dict["sigma_2"]]

            if self.name in {'Sum', 'RBF', 'Product'}:
                self.ll = np.array(
                    [self.array[self.key_dict["ll"]]] * self.D
                )
            if self.name in {"RBF", "Sum", "RBF_ARD"}:
                self.sigma_rbf = self.array[
                    self.key_dict["sigma_rbf"]
                ]
            if self.name == "RBF_ARD":
                self.ll = np.array(
                    self.array[self.key_dict["ll"]:self.length]
                )
            if self.name == 'Product':
                self.sigma_hat = self.array[
                    self.key_dict["sigma_hat"]
                ]
            if self.name in {'Sum', 'Directed', 'Product'}:
                gamma = self.array[self.key_dict["direc"]]
                self.direc = np.array([np.cos(gamma), np.sin(gamma)])
                if self.name != 'Product':
                    self.sigma_dir = self.array[
                        self.key_dict["sigma_dir"]
                    ]
                self.ll_dir = self.array[self.key_dict["ll_dir"]]
                self.B_0 = 1 - self.direc[0] ** 2
                self.B_1 = 1 - self.direc[1] ** 2
                self.B_2 = -self.direc[0] * self.direc[1]
                self.A = np.array(
                    [[self.B_0, self.B_2], [self.B_2, self.B_1]]
                )
            if self.name == 'Product':
                ll_ratio = self.ll_dir**2/self.ll[0]**2
                self.E_0 = ll_ratio + self.B_0
                self.E_1 = ll_ratio + self.B_1
                self.E_2 = self.B_2
                self.E = np.array(
                    [[self.E_0, self.E_2], [self.E_2, self.E_1]]
                )
                
                
    def get_val(self, key, kk=None, jj=None):
        if self.name in ['Combination', 'Temporal']:
            try:
                return self.array[self.key_dict[kk][key]]
            except:
                return self.array[self.key_dict[kk][jj][key]]
        else:
            return self.array[self.key_dict[key]]
