import sys
import argparse
import pickle
import copy

print('Starting to write experiment file')

def parse_arguments():
    parser = argparse.ArgumentParser()
    
    parser.add_argument("seed_max", help="max seed for numpy.random.", type=int)
    parser.add_argument("--file_min", type=int, default=None)
    parser.add_argument("--file_max", type=int, default=None)
    parser.add_argument("sID", help="student ID.", type=str)
    parser.add_argument("out_file", help="experiment file to write to.", type=str)
    parser.add_argument("--seed_min", help="min seed for numpy.random.", type=int,
                       default=0)
    parser.add_argument("-Q", "--iterations", help="number of BO iterations.", 
                        type=int, default=10)
    
    parser.add_argument("--sigma_2", help="hyperparameter default for sigma_2.", nargs='+', type=float, default=None)
    parser.add_argument("--direc", help="hyperparameter default for direc.", nargs='+', type=float, default=None)
    parser.add_argument("--sigma_dir", help="hyperparameter default for sigma_dir.", nargs='+', type=float, default=None)
    parser.add_argument("--sigma_rbf", help="hyperparameter default for sigma_rbf.", nargs='+', type=float, default=None)
    parser.add_argument("--sigma_hat", help="hyperparameter default for sigma_hat.", nargs='+', type=float, default=None)
    parser.add_argument("--ll_dir", help="hyperparameter default for ll_dir.", nargs='+', type=float, default=None)
    parser.add_argument("--ll", help="hyperparameter default for ll.", nargs='+', type=float, default=None)
    parser.add_argument("--prob_kernel", help="kernel type for setting up problem.", type=str, default=None)
    parser.add_argument("--sol_kernel", help="kernel type for BO.", type=str, default=None)
    parser.add_argument("--append", help="whether to append to the file.", type=bool, default=False)
    parser.add_argument("--ll_dir_prob", help="directed lengthscale of problem.", type=float, default=None)
    parser.add_argument("--ll_prob", help="rbf lengthscale of problem.", type=float, default=None)
    parser.add_argument("--dom_size", help="domain size.", type=int, default=None)
    parser.add_argument("--sigma_dir_prob", help="directed signal variance of problem.", type=float, default=None)
    parser.add_argument("--sigma_rbf_prob", help="rbf signal variance of problem.", type=float, default=None)
    parser.add_argument("--sigma_hat_prob", help="signal variance of product problem.", type=float, default=None)
    parser.add_argument("--acq", help="acquisition function to use.", type=str, default=None)
    parser.add_argument("--prob_file", help="file to load problem from.", type=str, default=None)
    parser.add_argument("--bounds_file", help="file to load bounds from. Overrides.", type=str, default=None)
    parser.add_argument("--initialisations", help="number of initialisations per optimisation run.", type=int, default=1)
    parser.add_argument("--sigma_2_prior", help="hyperparameter prior for sigma_2.", nargs='+', type=float, default=None)
    parser.add_argument("--sigma_dir_prior", help="hyperparameter prior for sigma_dir.", nargs='+', type=float, default=None)
    parser.add_argument("--sigma_rbf_prior", help="hyperparameter prior for sigma_rbf.", nargs='+', type=float, default=None)
    parser.add_argument("--sigma_hat_prior", help="hyperparameter prior for sigma_hat.", nargs='+', type=float, default=None)
    parser.add_argument("--ll_dir_prior", help="hyperparameter prior for ll_dir.", nargs='+', type=float, default=None)
    parser.add_argument("--ll_prior", help="hyperparameter prior for ll.", nargs='+', type=float, default=None)
    parser.add_argument("--ucb_beta", help="beta parameter for ucb acquisition function.", default=None)
    parser.add_argument("--priors_file", help="file to load priors from. Overrides.", type=str, default=None)
    parser.add_argument("--priors_std_bias", help="which std file to use.", type=bool, default=True)

    
    args = parser.parse_args()
    
    return args

args = parse_arguments()

if args.append:
    write_mode = "a"
else:
    write_mode = "w"

out_file = open(args.out_file, write_mode)
Q = args.iterations
sID = args.sID
task_seed_max = args.seed_max

exp_script = "/home/%s/BO/test_run_cluster.py" %sID

if args.bounds_file != None:
    print('Loading bounds dict from %s' %args.bounds_file)
    bounds_fl_dict = pickle.load(open(args.bounds_file, 'rb'))
    
    key = "sigma_2"
    if key in bounds_fl_dict.keys():
        if args.sigma_2 is not None:
            print('bound %s overwritten.' %key)
        args.sigma_2 = list(bounds_fl_dict[key])
        print('bound %s \t loaded from file.' %key)
    key = "direc"
    if key in bounds_fl_dict.keys():
        if args.direc is not None:
            print('bound %s overwritten.' %key)
        args.direc = list(bounds_fl_dict[key])
        print('bound %s \t loaded from file.' %key)
    key = "sigma_dir"
    if key in bounds_fl_dict.keys():
        if args.sigma_dir is not None:
            print('bound %s overwritten.' %key)
        args.sigma_dir = list(bounds_fl_dict[key])
        print('bound %s \t loaded from file.' %key)
    key = "sigma_rbf"
    if key in bounds_fl_dict.keys():
        if args.sigma_rbf is not None:
            print('bound %s overwritten.' %key)
        args.sigma_rbf = list(bounds_fl_dict[key])
        print('bound %s \t loaded from file.' %key)
    key = "sigma_hat"
    if key in bounds_fl_dict.keys():
        if args.sigma_hat is not None:
            print('bound %s overwritten.' %key)
        args.sigma_hat = list(bounds_fl_dict[key])
        print('bound %s \t loaded from file.' %key)
    key = "ll"
    if key in bounds_fl_dict.keys():
        if args.ll is not None:
            print('bound %s overwritten.' %key)
        args.ll = list(bounds_fl_dict[key])
        print('bound %s \t loaded from file.' %key)
    key = "ll_dir"
    if key in bounds_fl_dict.keys():
        if args.ll_dir is not None:
            print('bound %s overwritten.' %key)
        args.ll_dir = list(bounds_fl_dict[key])
        print('bound %s \t loaded from file.' %key)
        
if args.priors_file != None:
    print('Loading priors dict from %s' %args.priors_file)
    priors_mean_dict = pickle.load(open(args.priors_file+'_mean.p', 'rb'))
    if args.priors_std_bias:
        priors_std_dict = pickle.load(open(args.priors_file+'_N_std.p', 'rb'))
    else:
        priors_std_dict = pickle.load(open(args.priors_file+'_N1_std.p', 'rb'))

    
    key = "sigma_2"
    if key in priors_mean_dict.keys():
        if args.sigma_2_prior is not None:
            print('prior %s overwritten.' %key)
        args.sigma_2_prior = list([priors_mean_dict[key], priors_std_dict[key]])
        print('prior %s \t loaded from file.' %key)
    key = "sigma_dir"
    if key in priors_mean_dict.keys():
        if args.sigma_dir_prior is not None:
            print('prior %s overwritten.' %key)
        args.sigma_dir_prior = list([priors_mean_dict[key], priors_std_dict[key]])
        print('prior %s \t loaded from file.' %key)
    key = "sigma_rbf"
    if key in priors_mean_dict.keys():
        if args.sigma_rbf_prior is not None:
            print('prior %s overwritten.' %key)
        args.sigma_rbf_prior = list([priors_mean_dict[key], priors_std_dict[key]])
        print('prior %s \t loaded from file.' %key)
    key = "sigma_hat"
    if key in priors_mean_dict.keys():
        if args.sigma_hat_prior is not None:
            print('prior %s overwritten.' %key)
        args.sigma_hat_prior = list([priors_mean_dict[key], priors_std_dict[key]])
        print('prior %s \t loaded from file.' %key)
    key = "ll"
    if key in priors_mean_dict.keys():
        if args.ll_prior is not None:
            print('prior %s overwritten.' %key)
        args.ll_prior = list([priors_mean_dict[key], priors_std_dict[key]])
        print('prior %s \t loaded from file.' %key)
    key = "ll_dir"
    if key in priors_mean_dict.keys():
        if args.ll_dir_prior is not None:
            print('prior %s overwritten.' %key)
        args.ll_dir_prior = list([priors_mean_dict[key], priors_std_dict[key]])
        print('prior %s \t loaded from file.' %key)
        


for ii in range(args.seed_min, task_seed_max):
    cur_str = "python %s %s -Q %s" %(exp_script, ii, Q)
    for par_str, par in [("sigma_2", args.sigma_2),
            ("direc", args.direc), 
            ("sigma_dir", args.sigma_dir),
            ("sigma_rbf", args.sigma_rbf),
            ("sigma_hat", args.sigma_hat),
            ("ll", args.ll),
            ("ll_dir", args.ll_dir),
            ("sigma_2_prior", args.sigma_2_prior),
            ("sigma_dir_prior", args.sigma_dir_prior),
            ("sigma_rbf_prior", args.sigma_rbf_prior),
            ("ll_prior", args.ll_prior),
            ("ll_dir_prior", args.ll_dir_prior),
            ("sigma_hat_prior", args.sigma_hat_prior)
           ]:
        if par is not None:
            cur_str = cur_str + " --%s %s %s" %(par_str, par[0], par[1])
    if args.file_min is None or args.file_max is None:
        for par_str, par in [("prob_kernel", args.prob_kernel),
                ("sol_kernel", args.sol_kernel),
                ("ll_dir_prob", args.ll_dir_prob),
                ("ll_prob", args.ll_prob),
                ("dom_size", args.dom_size),
                ("sigma_dir_prob", args.sigma_dir_prob),
                ("sigma_rbf_prob", args.sigma_rbf_prob),
                ("sigma_hat_prob", args.sigma_hat_prob),
                ("acq", args.acq),
                ("prob_file", args.prob_file),
                ("initialisations", args.initialisations),
                ("ucb_beta", args.ucb_beta)
                ]:
            if par is not None:
                cur_str = cur_str + " --%s %s" %(par_str, par)
        print(cur_str, file=out_file)
    else:
        cur_str_raw = copy.deepcopy(cur_str)
        for jj in range(args.file_min, args.file_max+1):
            cur_str = copy.deepcopy(cur_str_raw)
            for par_str, par in [("prob_kernel", args.prob_kernel),
                    ("sol_kernel", args.sol_kernel),
                    ("ll_dir_prob", args.ll_dir_prob),
                    ("ll_prob", args.ll_prob),
                    ("dom_size", args.dom_size),
                    ("sigma_dir_prob", args.sigma_dir_prob),
                    ("sigma_rbf_prob", args.sigma_rbf_prob),
                    ("sigma_hat_prob", args.sigma_hat_prob),
                    ("acq", args.acq),
                    ("prob_file", '%s_%s.p' %(args.prob_file, jj)),
                    ("initialisations", args.initialisations),
                    ("ucb_beta", args.ucb_beta)
                    ]:
                if par is not None:
                    cur_str = cur_str + " --%s %s" %(par_str, par)

            print(cur_str, file=out_file)
out_file.close()
