import pickle
import matplotlib.pyplot as plt
import numpy as np
import datetime
import sys

sys.path.append('../BO')

from source import setup_helper

poll = 'NO2' 
tt = 40 
classes = ['Roadside']
classes_str = '_'.join(classes)
year = 2017

data = pickle.load(open(
    'data_sorted/%s_year_%s_ratified.p' %(poll, year), 'rb'))
clss_dict = pickle.load(open(
    'classification_dict.p', 'rb'))

loc_dict = pickle.load(open('lat_lon_dict.p', 'rb'))
lat_min, lon_min = np.inf, np.inf
for key in loc_dict.keys():
    lat, lon = loc_dict[key]
    if lat is not None and lon is not None:
        lat_min = min(lat_min, lat)
        lon_min = min(lon_min, lon)

def convert_lat_lon_km(domain, min_lat, min_lon):
    R = 6371
    lat = domain[:,0]
    lon = domain[:,1]
    x1 = (lat - min_lat) * 111
    x0 = (lon - min_lon) * np.pi/180*R*np.cos(np.pi/180*lat)
    return np.vstack([x1, x0]).T
      
def add_val(val, dic, key):
    if key in dic.keys():
        dic[key].append(val)
    else:
        dic[key] = [val]
        
days_data = {}
for dd in data:
    if clss_dict[dd[0]] in classes:
        day = (dd[1] - datetime.datetime(year-1, 12, 31)).days
        add_val([dd[2], dd[4], dd[5]], days_data, day)
    
for key in days_data.keys():
    days_data[key] = np.array(days_data[key])
    
print([(key, len(days_data[key])) for key in days_data.keys()])
    
cand = [ii for ii in range(1, 366) if len(days_data[ii]) >= tt]
np.random.seed(13)
for ii in cand:
    dat = days_data[ii]
    domain = convert_lat_lon_km(dat[:,[1, 2]], lat_min, lon_min)
    true_func = dat[:,0]
    idx = np.random.randint(0, len(dat), 5)
    xx = domain[idx,:]
    yy = true_func[idx]
    
    prob = setup_helper.Problem_obj(
        xx_nn=xx,
        yy_nn=yy,
        domain=domain,
        true_func=true_func,
        kernel=None,
        hyp_dict=None,
        C=None,
        dom_1D_vecs=None,
        seed='%s-%s-%s' %(poll, year, ii),
        sampling_noise=10e-8)
    
    pickle.dump(
        prob, 
        open('data_sorted/%s_problems/spatial_%s_%s_km_tt_%s_day-%s.p' %(year, poll, classes_str, tt, ii), 'wb')) 
    
    log_true_func = np.log(true_func)
    log_yy = np.log(yy)
    
    prob = setup_helper.Problem_obj(
        xx_nn=xx,
        yy_nn=log_yy - np.mean(log_true_func),
        domain=domain,
        true_func=log_true_func - np.mean(log_true_func),
        kernel=None,
        hyp_dict=None,
        C=None,
        dom_1D_vecs=None,
        seed='(%s-%s-%s' %(poll, year, ii),
        sampling_noise=10e-8)
    
    pickle.dump(
        prob, 
        open('data_sorted/%s_problems/spatial_log_mean_%s_%s_km_tt_%s_day-%s.p' %(year, poll, classes_str, tt, ii), 'wb')) 
