#!/bin/sh
#SBATCH -N 1                # Nodes requested
#SBATCH -n 1                # Tasks requested
#SBATCH --cpus-per-task=8   # CPUs requested
#SBATCH --mem=14000         # Memory in Mb
#SBATCH --time=2-12:00:00    # Time required
#SBATCH --exclude=REDACTED
#SBATCH -o ./Log_files/slurm-%A_%a.out
export STUDENT_ID=$(whoami)

export PYTHON_PATH=$PATH

#mkdir -p /home/${STUDENT_ID}/BO/Report/${SLURM_ARRAY_JOB_ID}

mkdir -p /disk/scratch/${STUDENT_ID}

JOBTASK="${SLURM_ARRAY_JOB_ID}-${SLURM_ARRAY_TASK_ID}"

export TMPDIR=/disk/scratch/${STUDENT_ID}/
export TMP=/disk/scratch/${STUDENT_ID}
export RESDIR=/disk/scratch/${STUDENT_ID}/${JOBTASK}

source /home/${STUDENT_ID}/miniconda3/bin/activate CONDA_ENVIRONMENT

echo "Machine:"
hostname
echo "Number of CPUS: "${SLURM_CPUS_PER_TASK}
TIMESTAMP=`date +%Y-%m-%d_%H-%M-%S`
echo "Timestamp: "${TIMESTAMP}
git log -1
git branch
conda info | grep active

experiment_text_file=$1
COMMAND="`sed \"${SLURM_ARRAY_TASK_ID}q;d\" ${experiment_text_file}`"

COMMAND_FULL="${COMMAND} -folder ${JOBTASK}"

# Synchronise data files
if [ ! -z "$2" ]; then
    data_locs=${2}
    echo ${data_locs}
    DATADIR=/home/${STUDENT_ID}/BO/$5 
    DATADIR_NODE=/disk/scratch/${STUDENT_ID}/DATA/
    eval "mkdir ${DATADIR_NODE}"
    zip_command="zip -j ${DATADIR}tmp_data_${JOBTASK}.zip ${DATADIR}/${data_locs}*.p"
    eval "${zip_command}"
    rsync -avuzh ${DATADIR}tmp_data_${JOBTASK}.zip ${DATADIR_NODE}tmp_data_${JOBTASK}.zip
    eval "unzip ${DATADIR_NODE}tmp_data_${JOBTASK}.zip -d ${DATADIR_NODE}"

    COMMAND_FULL="${COMMAND_FULL} -data_locs ${DATADIR_NODE}${data_locs} $3 $4"
fi

# Synchronise kernel file
if [ ! -z "$6" ]; then
    kernel_folder=kernels/
    rsync -avuzh ${kernel_folder}$6 ${DATADIR_NODE}$6
    COMMAND_FULL="${COMMAND_FULL} -kernel_hyp_file ${DATADIR_NODE}$6"
fi


echo "Running provided command: ${COMMAND_FULL}"
eval "${COMMAND_FULL}"
echo "Command ran successfully!"

# Extract result files
cd ${RESDIR}
if [ $? -eq 0 ]; then
    zip -rDo ${TMPDIR}${JOBTASK}.zip *

    OUT_FOLDER=/home/${STUDENT_ID}/MCMC/Hierarchical/Log_files/${SLURM_ARRAY_JOB_ID}
    echo "Sending results to: ${OUT_FOLDER}/${JOBTASK}"
    mkdir -p ${OUT_FOLDER}

    rsync -avuzh ${TMPDIR}${JOBTASK}.zip ${OUT_FOLDER}/${JOBTASK}.zip
fi

eval "rm ${DATADIR}tmp_data_${JOBTASK}.zip"


# Check whether to send email stating completion
e_file=${JOBTASK}_email.txt
exp_id=${SLURM_ARRAY_JOB_ID}
echo "Subject: Experiment ${exp_id} finished" > ${e_file}
echo "The experiment has finished running on the cluster." >> ${e_file}
date >> ${e_file}
cat ${e_file}
echo ${exp_id} >> ${e_file}
if [ "$(squeue | grep -c ${exp_id})" -ge 2 ]
then
    echo 'Still running'
else
    echo 'Experiment finished'
    sendmail REDACTED < ${e_file}
fi
