

import scipy.stats as st
import numpy as np
import copy

import sys
try:
    sys.path.append('../../BO/source/')
    import hyp_dict as hd
except:
    sys.path.append('/home/REDACTED/BO/source/')
    import hyp_dict as hd
    
import setup_helper
import kernel_functions_mul
import hyp_dict as hd


class Problem():
    
    def calc_log_likelihood(self, proposal, xx_grid, sample):
        raise NotImplementedError
    
#     class gamma_vars():
#         pass
    class eta_vars():
        def propose(self):
            NotImplementedError
        def prior(self):
            NotImplementedError
   
    
    def pdf_theta_log_multiple(self, theta, eta):
        tally = 0
        for ii in range(len(theta)):
            tally += self.pdf_theta_log(theta[ii], eta)
        return tally
    
    class theta_vars():
        # Structure to hold variables of eta: the hyperparameters of the RBF kerenl:
        # lengthscale and signal variance
        def __init__(self, theta_vv):
            self.vars = theta_vv
            self.order = sorted(theta_vv.keys())
        def array(self):
            return [self.vars[key] for key in self.order]
        
    def sample_theta(self, eta):
        # Sample theta | eta
        # Independently sample from gamma distributions defined by eta
        sampled = {key: eta.vars[key].sample() 
                   for key in eta.order}

        return self.theta_vars(sampled)

    def pdf_theta_log(self, theta, eta):
        # log pdf of theta given a value of eta. Used to calculate the likelihood of eta.
        for key in eta.vars.keys():
            if not eta.vars[key].valid():
                return 0
        tally = 0
        for key in eta.vars.keys():
            factor = eta.vars[key].logpdf(theta.vars[key])
            tally += factor
            if factor < -1000:
                print('unlikely %s: \t%s\t%s\t%s' %(
                key, theta.vars[key], factor, eta.vars[key].vals
                ))
        return tally

    
    def theta_part_pdf(self, theta_part, eta, part):
        if part is None:
            if not eta.valid():
                return np.zeros(len(theta_part))
            return eta.pdf(theta_part)
        
        if not eta.vars[part].valid():
            return np.zeros(len(theta_part))
        return eta.vars[part].pdf(theta_part)
    
    
class gamma_vars():
    # Structure to hold parameters defining a gamma distribution
    def __init__(self, shape, scale):
        self.shape = shape
        self.scale = scale
        self.len = 2
            
    def sample(self):
        ii = 0
        samp = np.random.gamma(shape=self.shape,
                scale=self.scale)
        while samp == 0 and ii < 1000:
            samp = np.random.gamma(shape=self.shape,
                scale=self.scale)
            ii += 1
        return samp
    def logpdf(self, value):
        return st.gamma.logpdf(value, a=self.shape, 
                scale=self.scale)
    def pdf(self, value):
        return st.gamma.pdf(value, a=self.shape, 
                scale=self.scale)
    def valid(self):
        if self.shape > 0 and self.scale > 0:
            return 1
        return 0
    def vals(self):
        return [self.shape, self.scale]
    def diff_copy(self, vals):
        return gamma_vars(vals[0], vals[1])
    def __str__(self):
        return 'shape: %s\tscale: %s' %(self.shape, self.scale)
    
    
class uniform_vars():
    # Structure to hold parameters defining a gamma distribution
    def __init__(self, low_b, high_b):
        self.low_b = low_b
        self.high_b = high_b
        self.len = 2
            
    def sample(self):
        return np.random.uniform(low=self.low_b, 
                                 high=self.high_b)
    def logpdf(self, value):
        return st.uniform.logpdf(value, loc=self.low_b, 
                scale=(self.high_b-self.low_b))
    def pdf(self, value):
        return st.uniform.pdf(value, loc=self.low_b, 
                scale=(self.high_b-self.low_b))
    def valid(self):
        if self.high_b >= self.low_b:
            return 1
        return 0
    def vals(self):
        return [self.low_b, self.high_b]
    def diff_copy(self, vals):
        return uniform_vars(vals[0], vals[1])
    def __str__(self):
        return 'low: %s\thigh: %s' %(self.low_b, self.high_b)
    

        
        
class Clean_problem(Problem):


    def calc_log_likelihood(self, proposal, xx_grid, sample):
        Nx = len(xx_grid)
        Ky = self.calc_K(xx_grid, proposal=proposal)

        return st.multivariate_normal.logpdf(
                sample, np.zeros(Nx), Ky)

    def calc_K(self, xx_grid, proposal):
        # calculates the covariance matrix for the RBF kernel given theta
        lscale = proposal.vars['lscale']
        sigma_n = proposal.vars['sigma_n']
        sigma_r = proposal.vars['sigma_r']
        
        Nx, D = np.shape(xx_grid)
        broadcast = np.broadcast_to(xx_grid.T, (Nx, D, Nx))
        broadcast_tpose = np.transpose(broadcast, [0, 2, 1])
        diff = broadcast_tpose - \
                np.transpose(broadcast_tpose, [1, 0, 2])
        Kxx = sigma_r*np.exp(-np.sum(diff**2/lscale**2, 2))

        return Kxx + np.eye(Nx)*(sigma_n**2+0.00001)
    
            
    class eta_vars():
    # Structure to hold variables of eta: the parameters defining two gamma distributions
    # on the lengthscale and signal variance of an RBF kernel
        def __init__(self, eta_vv, args, constant=[]):
            self.set_vars(eta_vv)
            self.order = sorted(eta_vv.keys())
            self.constant = constant
            
            sigma_e = []
            for key in self.order:
                if key in constant:
                    for _ in range(self.vars[key].len):
                        sigma_e.append(0)
                else:
                    if key in ['sigma_rbf', 'sigma_hat']:
                        key = 'sigma_r'
                    if key == 'lscale':
                        key = 'll'
                    if args.kernel == 'Combination' and key != 'sigma_n':
                        props = args.__dict__['C'+key+'_prop']
                    else:
                        props = args.__dict__[key+'_prop']
                    for val in props:
                        sigma_e.append(val)
            self.sigma_e = np.array(sigma_e)
            print('sigma_e set to %s' %self.sigma_e)
            const_idx = []
            ii = 0
            for key in self.order:
                if key in constant:
                    const_idx += range(ii, ii+self.vars[key].len)

                ii += self.vars[key].len
            self.const_idx = const_idx


        def set_vars(self, eta_vv):
            self.vars = eta_vv

        def array(self):
            # Return an array of the varying parameters (the scale is set constant)
            return np.hstack([
                self.vars[var].vals() for var in self.order])

        def propose_part(self, ii):
            proposal_array = copy.deepcopy(self.array())
            if ii in self.const_idx:
                return self.eta_vars_array(proposal_array)
            proposal_array[ii] = np.random.normal(
                proposal_array[ii], self.sigma_e[ii])
            
            return self.eta_vars_array(proposal_array)
        
        def propose(self):
            # Proposal distribution used for sampling eta|theta
            array = self.array()
            proposal_array = np.random.multivariate_normal(
                array, np.eye(len(array))*self.sigma_e)
            return self.eta_vars_array(proposal_array)
        
        def eta_vars_array(self, array):
            # Build eta structure from array

            built = {}
            kk = 0
            for key in self.order:
                built[key] = self.vars[key].diff_copy(
                    array[kk:kk+self.vars[key].len])
                kk += self.vars[key].len
            copy_self = copy.deepcopy(self)
            copy_self.set_vars(built)
            return copy_self 
        
        def prior(self):
            if np.all(self.array() >= 0):
                return 1
            else:
                return 0
            
        def valid(self):
            for key in self.vars.keys():
                if not self.vars[key].valid():
                    return 0
            return 1
        


            
class Multiple_problem(Clean_problem):
        
    def calc_log_likelihood(self, proposal, xx_grid, sample):
        tally = 0
        for (x_loc, samp) in zip(xx_grid, sample):
            Nx = len(x_loc)
            Ky = self.calc_K(x_loc, proposal=proposal)
            try:
                tally += st.multivariate_normal.logpdf(
                    samp, np.zeros(Nx), Ky)
            except:
                try:
                    D, _ = np.shape(Ky)
                    tally += -D/2*np.log(2*np.pi) \
                        - 1/2 * np.log(np.linalg.det(Ky)) \
                        - 1/2 * np.matmul(
                            np.matmul(samp.T, np.linalg.inv(Ky)),
                            samp)
                    print('Error in logpdf calculation catched.')
                    
                except:
                    print('Error: %s, %s, %s' %(
                        proposal.lscale, proposal.sigma_n, proposal.sigma_r
                    ))
                    raise ValueError
        return tally
    
    
class Multiple_kernel_problem(Multiple_problem):
    
    def __init__(self, kernel_name, parts=None, 
                 hyp_dict=None, kernel=None):
        if kernel is None:
            self.kernel = kernel_functions_mul.choose_kernel(
                kernel_name, parts=parts)
        else:
            self.kernel = kernel
        self.parts = parts
        self.hyp_dict = hyp_dict
        
    def build_hyp_dict(self, theta):
        raw_dict = {'kernel_name': self.kernel.name}
        for key in theta.vars.keys():
            if key == 'sigma_n':
                if self.kernel.name in ['Combination', 'Temporal']:
                    if type(self.parts[0]) is not list:
                        raw_dict['0_sigma_2'] = theta.vars[key]**2
                    else:
                        raw_dict['0_0_sigma_2'] = theta.vars[key]**2
                else:
                    raw_dict['sigma_2'] = theta.vars[key]**2
            else:
                raw_dict[key] = theta.vars[key]
        if self.hyp_dict is None:
            return hd.Hyp_dict(raw_dict, parts=self.parts)
        else:
            self.hyp_dict.update_vals_from_dict(raw_dict)
            return self.hyp_dict
    
    def calc_K(self, x_loc, proposal=None):
        hyp_dict = self.build_hyp_dict(proposal)
        return self.kernel.knn_with_noise(
                domain=x_loc, hyp_dict=hyp_dict)
    
    def get_test_image(self, N, eta, domain,
                   N_start=2
                  ):
        
        sample = self.sample_theta(eta)
        
        hyp_dict = self.build_hyp_dict(sample)

        test_y = np.random.multivariate_normal(
            np.zeros(N), self.kernel.knn_with_noise(
                domain=domain, hyp_dict=hyp_dict))

        start_idx = np.random.randint(0, N, N_start)
        Y = test_y[start_idx]
        X = domain[start_idx]
        vmin = np.min(test_y)
        vmax = np.max(test_y)

        prob = setup_helper.Problem_obj(
            xx_nn=X,
            yy_nn=Y,
            domain=domain,
            true_func=test_y,
            kernel=self.kernel,
            hyp_dict=hyp_dict,
            C=None,
            dom_1D_vecs=np.array([np.array(range(28))]*2),
            seed=None,
            sampling_noise=0
        )

        correct_hyp = {'len': 1}
        for key in sample.vars.keys():
            if key == 'sigma_n':
                correct_hyp['sigma_2'] = [sample.vars[key]**2]
            else:
                correct_hyp[key] = [sample.vars[key]]

        return prob, correct_hyp
            
            
def get_test_image(N, lscale_rbf_gamma, sigma_rbf_gamma,
                   sigma_noise_gamma, domain,
                   kernel,
                   sigma_key,
                   N_start=2,
                   lscale_dir_gamma=None
                  ):
    lambda_r_test = lscale_rbf_gamma.sample()

    sigma_r_test = sigma_rbf_gamma.sample()

    sigma_n_test = sigma_noise_gamma.sample()

    if kernel.name in ['Directed', 'Sum', 'Product']:
        lambda_d_test = lscale_dir_gamma.sample()

    dir_test = np.random.uniform(0, np.pi)
    
    raw_hyp_dict = {
            'kernel_name': kernel.name,
            'll': lambda_r_test,
            'sigma_2': sigma_n_test**2,
            sigma_key: sigma_r_test,
            'direc': dir_test
        }
    if kernel.name in ['Directed', 'Sum', 'Product']:
        raw_hyp_dict['ll_dir'] = lambda_d_test
    hyp_dict = hd.Hyp_dict(raw_hyp_dict)
    
    test_y = np.random.multivariate_normal(
        np.zeros(N), kernel.knn_with_noise(
            domain=domain, hyp_dict=hyp_dict))
    
    start_idx = np.random.randint(0, N, N_start)
    Y = test_y[start_idx]
    X = domain[start_idx]
    
    prob = setup_helper.Problem_obj(
        xx_nn=X,
        yy_nn=Y,
        domain=domain,
        true_func=test_y,
        kernel=kernel,
        hyp_dict=hyp_dict,
        C=None,
        dom_1D_vecs=np.array([np.array(range(28))]*2),
        seed=None,
        sampling_noise=0
    )
    
    correct_hyp = {
        'll': [lambda_r_test],
        'sigma_2': [sigma_n_test**2],
        sigma_key: [sigma_r_test],
        'direc': [dir_test],
        'len': 1
    }
    if kernel.name in ['Directed', 'Sum', 'Product']:
        correct_hyp['ll_dir'] = [lambda_d_test]
    
    return prob, correct_hyp
