import pickle
import numpy as np

import problems
import argparse

def parse_arguments():
    parser = argparse.ArgumentParser()
    
    parser.add_argument("-seed", help="seed for numpy.random.", type=int, default=3)
    parser.add_argument("-N", help="number of samples to use.", type=int, 
                        default=10)
    parser.add_argument("-timestamp", help="timestamp of run to use.", type=str)
    parser.add_argument('-burn_in', type=int, default=None, help='Overrides parameter from input file.')
    args = parser.parse_args()
    
    return args

args = parse_arguments()

seed = args.seed
np.random.seed(seed)
num_samples = args.N
timestamp = args.timestamp
file_in = 'Log_files/result_file_%s.p' %timestamp
file_out = 'Hyp_samples/%s_size_%s.p' %(timestamp, num_samples)

res_struct = pickle.load(open(file_in, 'rb'))

if args.burn_in is None:
    burn_in = res_struct['args'].burn_in
else:
    burn_in = args.burn_in

ee_arrays = res_struct['ee_arrays']
samp_struct = {'args': res_struct['args'],
               'timestamp': timestamp}
hyp_dict = {'len': num_samples}

Nall, _ = np.shape(ee_arrays)
all_idx = list(range(Nall - burn_in))
np.random.shuffle(all_idx)
chos_idx = np.array(all_idx[:num_samples]) + burn_in 

def check_membership(ent, gamma_vars):
    for gg in gamma_vars:
        len_gg = len(gg)
        if ent[-len_gg:] == gg:
            return True
    return False

gamma_vars = ['ll', 'll_dir', 'sigma_hat', 'sigma_n', 
              'sigma_r', 'sigma_dir']
translate_dict = {'dir': 'direc'}
ii = 0
for ent in res_struct['order']:
    ent_key = translate_dict[ent] if ent in translate_dict.keys() else ent
    var_1 = ee_arrays[chos_idx, 2*ii]
    var_2 = ee_arrays[chos_idx, 2*ii+1]
    
    ent_list = []
    if check_membership(ent, gamma_vars):
        print('%s: \t Gamma' %ent)
    else:
        print('%s: \t Uniform' %ent)
    for jj in range(len(chos_idx)):
        if check_membership(ent, gamma_vars):

            gamma = problems.gamma_vars(shape=var_1[jj], scale=var_2[jj]) 
            sample = gamma.sample()
        else:
            uniform = problems.uniform_vars(low_b=var_1[jj], high_b=var_2[jj])
            sample = uniform.sample()
        if ent == 'sigma_n':
            sample = sample ** 2
        ent_list.append(sample)
    if ent == 'sigma_n':
        hyp_dict['sigma_2'] = ent_list
    elif ent[-9:] == 'sigma_hat' and samp_struct['args'].kernel == 'Sum':
        hyp_dict['sigma_rbf'] = ent_list
    else:
        hyp_dict[ent_key] = ent_list
    ii += 1

    
samp_struct['samples'] = hyp_dict
pickle.dump(samp_struct, open(file_out, 'wb'))
